@@ -2,18 +2,40 @@
 to be cleaned at some point. """
 import logging
 import os
+import pickle
+import docker.errors
 
-from typing import Dict
+from enum import Enum
+from typing import Dict, List, Union
 from pathlib import Path
 
 from vantage6.common.globals import APPNAME
+from vantage6.common.docker.addons import (
+    remove_container_if_exists, remove_container
+)
 from vantage6.node.util import logger_name
+from vantage6.node.globals import ALPINE_IMAGE
 from vantage6.node.docker.vpn_manager import VPNManager
-from vantage6.node.docker.network_manager import IsolatedNetworkManager
+from vantage6.common.docker.network_manager import NetworkManager
 from vantage6.node.docker.docker_base import DockerBaseManager
-from vantage6.node.docker.utils import running_in_docker, remove_container
-from vantage6.common.docker_addons import pull_if_newer
+from vantage6.common.docker.addons import pull_if_newer, running_in_docker
+from vantage6.node.docker.exceptions import (
+    UnknownAlgorithmStartFail,
+    PermanentAlgorithmStartFail
+)
 
+class TaskStatus(Enum):
+    # Task constructor is executed
+    INITIALIZED = 0
+    # Container started without exceptions
+    STARTED = 1
+    # Container exited and had zero exit code
+    # COMPLETED = 2
+    # Failed to start the container
+    FAILED = 90
+    PERMANENTLY_FAILED = 91
+    # Container had a non zero exit code
+    # CRASHED = 92
 
 class DockerTaskManager(DockerBaseManager):
     """
@@ -28,9 +50,9 @@ class DockerTaskManager(DockerBaseManager):
 
     def __init__(self, image: str, vpn_manager: VPNManager, node_name: str,
                  result_id: int, tasks_dir: Path,
-                 isolated_network_mgr: IsolatedNetworkManager,
-                 database_uri: str, database_is_file: bool,
-                 docker_volume_name: str):
+                 isolated_network_mgr: NetworkManager,
+                 databases: dict, docker_volume_name: str,
+                 alpine_image: Union[str, None] = None):
         """
         Initialization creates DockerTaskManager instance
 
@@ -46,27 +68,29 @@ class DockerTaskManager(DockerBaseManager):
             Server result identifier
         tasks_dir: Path
             Directory in which this task's data are stored
-        isolated_network_mgr: IsolatedNetworkManager
+        isolated_network_mgr: NetworkManager
             Manager of isolated network to which algorithm needs to connect
-        database_uri: str
-            Location of the database
-        database_is_file: bool
-            Whether or not the database is a file
+        databases: Dict
+            List of databases
         docker_volume_name: str
             Name of the docker volume
+        alpine_image: str or None
+            Name of alternative Alpine image to be used
         """
         super().__init__(isolated_network_mgr)
         self.image = image
         self.__vpn_manager = vpn_manager
         self.result_id = result_id
         self.__tasks_dir = tasks_dir
-        self.__database_uri = database_uri
-        self.database_is_file = database_is_file
+        self.databases = databases
         self.data_volume_name = docker_volume_name
         self.node_name = node_name
+        self.alpine_image = ALPINE_IMAGE if alpine_image is None \
+            else alpine_image
 
         self.container = None
         self.status_code = None
+        self.docker_input = None
 
         self.labels = {
             f"{APPNAME}-type": "algorithm",
@@ -81,6 +105,9 @@ class DockerTaskManager(DockerBaseManager):
         self.tmp_folder = "/mnt/tmp"
         self.data_folder = "/mnt/data"
 
+        # keep track of the task status
+        self.status: TaskStatus = TaskStatus.INITIALIZED
+
     def is_finished(self) -> bool:
         """
         Checks if algorithm container is finished
@@ -137,7 +164,7 @@ class DockerTaskManager(DockerBaseManager):
             self.log.error(e)
 
     def run(self, docker_input: bytes, tmp_vol_name: str, token: str,
-            algorithm_env: Dict) -> int:
+            algorithm_env: Dict, database: str) -> List[Dict]:
         """
         Runs the docker-image in detached mode.
 
@@ -157,32 +184,33 @@ class DockerTaskManager(DockerBaseManager):
 
         Returns
         -------
-        int or None:
-            Port number assigned for VPN communication. None if VPN is inactive
+        List[Dict] or None
+            Description of each port on the VPN client that forwards traffic to
+            the algo container. None if VPN is not set up.
         """
         # generate task folders
         self._make_task_folders()
 
         # prepare volumes
-        self.volumes = self._prepare_volumes(
-            docker_input, tmp_vol_name, token
-        )
+        self.docker_input = docker_input
+        self.volumes = self._prepare_volumes(tmp_vol_name, token)
         self.log.debug(f"volumes: {self.volumes}")
 
         # setup environment variables
         self.environment_variables = \
-            self._setup_environment_vars(algorithm_env=algorithm_env)
+            self._setup_environment_vars(algorithm_env=algorithm_env,
+                                         database=database)
 
         # run the algorithm as docker container
-        vpn_port = self._run_algorithm()
-        return vpn_port
+        vpn_ports = self._run_algorithm()
+        return vpn_ports
 
     def cleanup(self) -> None:
         """Cleanup the containers generated for this task"""
         remove_container(self.helper_container, kill=True)
         remove_container(self.container, kill=True)
 
-    def _run_algorithm(self) -> int:
+    def _run_algorithm(self) -> List[Dict]:
         """
         Run the algorithm container
 
@@ -191,16 +219,24 @@ class DockerTaskManager(DockerBaseManager):
 
         Returns
         -------
-        int or None:
-            Port number assigned for VPN communication. None if VPN is inactive
+        List[Dict] or None
+            Description of each port on the VPN client that forwards traffic to
+            the algo container. None if VPN is inactive
         """
-        vpn_port = None
+        vpn_ports = None
         container_name = f'{APPNAME}-{self.node_name}-result-{self.result_id}'
         helper_container_name = container_name + '-helper'
 
+        # Try to pull the latest image
+        self.pull()
+
         # remove algorithm containers if they were already running
-        self.remove_container_if_exists(name=container_name)
-        self.remove_container_if_exists(name=helper_container_name)
+        remove_container_if_exists(
+            docker_client=self.docker, name=container_name
+        )
+        remove_container_if_exists(
+            docker_client=self.docker, name=helper_container_name
+        )
 
         if self.__vpn_manager:
             # if VPN is active, network exceptions must be configured
@@ -209,7 +245,7 @@ class DockerTaskManager(DockerBaseManager):
             # will therefore also affect the algorithm.
             self.helper_container = self.docker.containers.run(
                 command='sleep infinity',
-                image='alpine',
+                image=self.alpine_image,
                 labels=self.helper_labels,
                 network=self.isolated_network_mgr.network_name,
                 name=helper_container_name,
@@ -217,17 +253,26 @@ class DockerTaskManager(DockerBaseManager):
             )
             # setup forwarding of traffic via VPN client to and from the
             # algorithm container:
-            vpn_port = self.__vpn_manager.forward_vpn_traffic(
+            vpn_ports = self.__vpn_manager.forward_vpn_traffic(
                 helper_container=self.helper_container,
                 algo_image_name=self.image
             )
 
-        # Try to pull the latest image
-        self.pull()
+        # try reading docker input
+        deserialized_input = None
+        if self.docker_input:
+            try:
+                deserialized_input = pickle.loads(self.docker_input)
+            except Exception:
+                pass
 
         # attempt to run the image
         try:
-            self.log.info(f"Run docker image={self.image}")
+            if deserialized_input:
+                self.log.info(f"Run docker image {self.image} with input "
+                              f"{deserialized_input}")
+            else:
+                self.log.info(f"Run docker image {self.image}")
             self.container = self.docker.containers.run(
                 self.image,
                 detach=True,
@@ -237,12 +282,18 @@ class DockerTaskManager(DockerBaseManager):
                 name=container_name,
                 labels=self.labels
             )
-        except Exception as e:
-            self.log.error('Could not run docker image!?')
-            self.log.error(e)
-            return None
 
-        return vpn_port
+        except docker.errors.ImageNotFound:
+            self.log.error(f'Could not find image: {self.image}')
+            self.status = TaskStatus.PERMANENTLY_FAILED
+            raise PermanentAlgorithmStartFail
+
+        except Exception as e:
+            self.status = TaskStatus.FAILED
+            raise UnknownAlgorithmStartFail(e)
+
+        self.status = TaskStatus.STARTED
+        return vpn_ports
 
     def _make_task_folders(self) -> None:
         """ Generate task folders """
@@ -265,15 +316,12 @@ class DockerTaskManager(DockerBaseManager):
         os.makedirs(self.task_folder_path, exist_ok=True)
         self.output_file = os.path.join(self.task_folder_path, "output")
 
-    def _prepare_volumes(self, docker_input: bytes, tmp_vol_name: str,
-                         token: str) -> Dict:
+    def _prepare_volumes(self, tmp_vol_name: str, token: str) -> Dict:
         """
         Generate docker volumes required to run the algorithm
 
         Parameters
         ----------
-        docker_input: bytes
-            Input that can be read by docker container
         tmp_vol_name: str
             Name of temporary docker volume assigned to the algorithm
         token: str
@@ -284,13 +332,13 @@ class DockerTaskManager(DockerBaseManager):
         Dict:
             Volumes to support running the algorithm
         """
-        if isinstance(docker_input, str):
-            docker_input = docker_input.encode('utf8')
+        if isinstance(self.docker_input, str):
+            self.docker_input = self.docker_input.encode('utf8')
 
         # Create I/O files & token for the algorithm container
         self.log.debug("prepare IO files in docker volume")
         io_files = [
-            ('input', docker_input),
+            ('input', self.docker_input),
             ('output', b''),
             ('token', token.encode("ascii")),
         ]
@@ -313,7 +361,8 @@ class DockerTaskManager(DockerBaseManager):
                 {"bind": self.data_folder, "mode": "rw"}
         return volumes
 
-    def _setup_environment_vars(self, algorithm_env: Dict = {}) -> Dict:
+    def _setup_environment_vars(self, algorithm_env: Dict = {},
+                                database: str = 'default') -> Dict:
         """"
         Set environment variables required to run the algorithm
 
@@ -351,11 +400,28 @@ class DockerTaskManager(DockerBaseManager):
             "PORT": os.environ.get("PROXY_SERVER_PORT", 8080),
             "API_PATH": "",
         }
-        if self.database_is_file:
-            environment_variables["DATABASE_URI"] = \
-                f"{self.data_folder}/{self.__database_uri}"
-        else:
-            environment_variables["DATABASE_URI"] = self.__database_uri
+
+        # Only prepend the data_folder is it is a file-based database
+        # This allows algorithms to access multiple data-sources at the
+        # same time
+        for label in self.databases:
+            db = self.databases[label]
+            var_name = f'{label.upper()}_DATABASE_URI'
+            environment_variables[var_name] = \
+                f"{self.data_folder}/{os.path.basename(db['uri'])}" \
+                if db['is_file'] else db['uri']
+
+        # Support legacy algorithms
+        try:
+            environment_variables["DATABASE_URI"] = (
+                f"{self.data_folder}/"
+                f"{os.path.basename(self.databases[database]['uri'])}"
+            )
+        except KeyError as e:
+            self.log.error(f"'{database}' database missing! This could crash "
+                           "legacy algorithms")
+            self.log.debug(e)
+
         self.log.debug(f"environment: {environment_variables}")
 
         # Load additional environment variables
